/*
 * Copyright 2011 NORDUnet A/S.  All rights reserved.
 * Copyright 2011 Red Hat, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "asn1/PA-ENC-TS-ENC.h"
#include "asn1/PA-OTP-ENC-REQUEST.h"
#include "asn1/PA-OTP-CHALLENGE.h"
#include "asn1/PA-OTP-REQUEST.h"

#include "asn1c_utils.h"
#include "codec.h"
#include "codec_json.h"
#include "internal.h"
#include "memory.h"
#include "plugins.h"

#include <krb5/preauth_plugin.h>

#include <string.h>
#include <errno.h>
#include <time.h>

#include <arpa/inet.h>

#include <verto-jsonrpc.h>

struct config_state {
  krb5_context context;
  krb5_kdcpreauth_callbacks cb;
  krb5_kdcpreauth_rock rock;
};

struct edata_state {
  krb5_kdcpreauth_edata_respond_fn respond;
  void *arg;
};

struct verify_state {
  krb5_kdcpreauth_verify_respond_fn respond;
  void *arg;
};

static krb5_preauthtype authhub_server_pa_type_list[] =
  { KRB5_PADATA_OTP_REQUEST, 0 };

static void
on_constraint_fail(void *app_key, struct asn_TYPE_descriptor_s *failed_type,
                   const void *failed_ptr, const char *errmsg_format, ...)
{
  va_list ap;
  char *tmp;

  va_start(ap, errmsg_format);
  if (vasprintf(&tmp, errmsg_format, ap) < 0)
  return;
  va_end(ap);

  DEBUGMSG(EINVAL, tmp);
  free(tmp);
}

static krb5_error_code
nonce_generate(krb5_context ctx, unsigned int length, char **buf,
               unsigned int *len)
{
  krb5_data tmp;
  krb5_error_code retval;
  krb5_timestamp time;

  retval = krb5_timeofday(ctx, &time);
  if (retval != 0)
    return retval;

  tmp.length = length + sizeof(time);
  tmp.data = (char *) malloc(tmp.length);
  if (!tmp.data)
    return ENOMEM;

  retval = krb5_c_random_make_octets(ctx, &tmp);
  if (retval != 0) {
    free(tmp.data);
    return retval;
  }

  *((krb5_timestamp*) tmp.data) = htonl(time);
  *buf = tmp.data;
  *len = tmp.length;
  return 0;
}

krb5_error_code
nonce_verify(krb5_context ctx, krb5_keyblock *armor_key, const char *data,
             unsigned int len)
{
  krb5_error_code retval = EINVAL;
  krb5_timestamp ts;
  PA_OTP_ENC_REQUEST_t *er = NULL;

  if (!ctx || !armor_key || !data)
    goto egress;

  /* Decode the PA-OTP-ENC-REQUEST structure */
  er = int_ber_decode(PA_OTP_ENC_REQUEST, data, len);
  if (!er)
    goto egress;

  /* Make sure the nonce is exactly the same size as the one generated */
  if (er->nonce.size != armor_key->length + sizeof(krb5_timestamp))
    goto egress;

  /* Check to make sure the timestamp at the beginning is still valid */
  ts = ntohl(((krb5_timestamp*) er->nonce.buf)[0]);
  retval = krb5_check_clockskew(ctx, ts);

egress:
  asn_DEF_PA_OTP_ENC_REQUEST.free_struct(&asn_DEF_PA_OTP_ENC_REQUEST, er, 0);
  return retval;
}

krb5_error_code
timestamp_verify(krb5_context ctx, const char *data, unsigned int len)
{
  krb5_error_code retval = EINVAL;
  PA_ENC_TS_ENC_t *et = NULL;
  krb5_timestamp time;
  struct tm zero(t);

  if (!ctx || !data)
    goto egress;

  /* Decode the PA-ENC-TS-ENC structure */
  et = int_ber_decode(PA_ENC_TS_ENC, data, len);
  if (!et)
    goto egress;

  /* Get the timestamp */
  asn_GT2time(&et->patimestamp, &t, 0);
  time = mktime(&t);
  if (time < 0)
    goto egress;

  /* Check the clockskew */
  retval = krb5_check_clockskew(ctx, time);

egress:
  asn_DEF_PA_ENC_TS_ENC.free_struct(&asn_DEF_PA_ENC_TS_ENC, et, 0);
  return retval;
}

static krb5_error_code
decrypt_encdata(krb5_context context, krb5_keyblock *armor_key,
                PA_OTP_REQUEST_t *req, char **data, unsigned int *len)
{
  krb5_error_code retval;
  krb5_enc_data encdata;
  krb5_data tmp;
  long int kvno = 0;

  if (!req)
    return EINVAL;

  tmp.length = req->encData.cipher.size;
  tmp.data = newa0(char, tmp.length);
  if (!tmp.data)
    return ENOMEM;

  asn_INTEGER2long(req->encData.kvno, &kvno);
  encdata.kvno = kvno;
  encdata.enctype = req->encData.etype;
  encdata.ciphertext.data = (char*) req->encData.cipher.buf;
  encdata.ciphertext.length = req->encData.cipher.size;
  retval = krb5_c_decrypt(context, armor_key, KRB5_KEYUSAGE_PA_OTP_REQUEST,
                          NULL, &encdata, &tmp);
  if (retval != 0) {
    DEBUGMSG(retval, "Unable to decrypt encData in PA-OTP-REQUEST.");
    free(tmp.data);
    return retval;
  }

  if (data)
    *data = tmp.data;
  else
    free(tmp.data);
  if (len)
    *len = tmp.length;

  return 0;
}

static krb5_error_code
authhub_server_init(krb5_context context, krb5_kdcpreauth_moddata *moddata_out,
                    const char **realmnames)
{
  return plugins_load(SOCKETDIR, (inventory**) moddata_out);
}

static void
authhub_server_fini(krb5_context context, krb5_kdcpreauth_moddata moddata)
{
  plugins_free((inventory*) moddata);
}

static int
authhub_server_flags(krb5_context context, krb5_preauthtype pa_type)
{
  return PA_HARDWARE | PA_REPLACES_KEY;
}

static json_object *
get_config(const char *name, const void *misc)
{
  const struct config_state *cfgs = (const struct config_state *) misc;
  char *tmp, *cfg;
  krb5_error_code retval;
  json_object *cnfg = NULL;

  if (asprintf(&tmp, "authhub:%s", name) < 0)
    return NULL;

  retval = cfgs->cb->get_string(cfgs->context, cfgs->rock, tmp, &cfg);
  free(tmp);
  if (retval)
    return NULL;

  if (cfg)
    cnfg = json_tokener_parse(cfg);
  free(cfg);
  return cnfg;
}

static void
finish_edata(PA_OTP_CHALLENGE_t *chl, void *misc)
{
  krb5_pa_data *pa = NULL;
  struct edata_state *eds = (struct edata_state *) misc;

  /* Verify our constraints */
  if (asn_DEF_PA_OTP_CHALLENGE.check_constraints(&asn_DEF_PA_OTP_CHALLENGE,
                                                 chl, on_constraint_fail,
                                                 NULL) < 0) {
    asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
    DEBUGMSG(0, "Unable to create well-formed challenge.");
    eds->respond(eds->arg, ENOMEM, NULL);
    return;
  }

#ifdef DEBUG
  char *debug = NULL;
  if (int_asn1c_dump(&asn_DEF_PA_OTP_CHALLENGE, chl, &debug)) {
    DEBUGMSG(0, "\n%s\n\n", debug);
    free(debug);
  }
#endif

  /* Build the output pa data */
  pa = new0(krb5_pa_data);
  if (pa) {
    pa->pa_type = KRB5_PADATA_OTP_CHALLENGE;
    if (!int_der_encode(&asn_DEF_PA_OTP_CHALLENGE, chl, (char**) &pa->contents,
                        &pa->length)) {
      DEBUGMSG(ENOMEM, "Unable to encode challenge.");
      free(pa);
      pa = NULL;
    }
  }

  asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
  eds->respond(eds->arg, pa ? 0 : ENOMEM, pa);
}

static void
authhub_server_get_edata(krb5_context context, krb5_kdc_req *request,
                         krb5_kdcpreauth_callbacks cb,
                         krb5_kdcpreauth_rock rock,
                         krb5_kdcpreauth_moddata moddata,
                         krb5_preauthtype pa_type,
                         krb5_kdcpreauth_edata_respond_fn respond, void *arg)
{
  krb5_error_code retval = -1;
  krb5_keyblock *armor_key = NULL;
  char *buf = NULL;
  PA_OTP_CHALLENGE_t *chl;
  unsigned int len;
  struct config_state cfgs;
  struct edata_state *eds;

  cfgs.context = context;
  cfgs.cb = cb;
  cfgs.rock = rock;

  /* Get the armor key.
   * This indicates the length of random data to use in the nonce. */
  armor_key = cb->fast_armor(context, rock);
  if (armor_key == NULL) {
    DEBUGMSG(ENOENT, "No armor key found when preparing challenge.");
    (*respond)(arg, ENOENT, NULL);
    return;
  }

  /* Allocate edata state */
  eds = new0(struct edata_state);
  if (!eds) {
    (*respond)(arg, ENOMEM, NULL);
    return;
  }
  eds->respond = respond;
  eds->arg = arg;

  /* Build the challenge object */
  chl = new0(PA_OTP_CHALLENGE_t);
  if (!chl) {
    free(eds);
    (*respond)(arg, ENOMEM, NULL);
    return;
  }

  /* Generate the nonce */
  retval = nonce_generate(context, armor_key->length, &buf, &len);
  if (retval != 0) {
    DEBUGMSG(retval, "Unable to generate nonce.");
    free(eds);
    asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
    (*respond)(arg, retval, NULL);
    return;
  }
  chl->nonce.buf = (unsigned char*) buf;
  chl->nonce.size = len;

  /* Call getTokenInfo */
  retval = plugins_getTokenInfo((inventory*) moddata,
                                cb->event_context(context, rock), chl,
                                get_config, finish_edata, &cfgs, eds, free);
  if (retval) {
    asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
    (*respond)(arg, retval, NULL);
  }
}

static void
finish_verify(bool verified, void *misc)
{
  struct verify_state *vstate = (struct verify_state*) misc;

  vstate->respond(vstate->arg,
                  verified ? 0 : KRB5KDC_ERR_PREAUTH_FAILED,
                  NULL, NULL, NULL);
}

static void
authhub_server_verify_padata(krb5_context context, krb5_data *req_pkt,
                             krb5_kdc_req *request,
                             krb5_enc_tkt_part *enc_tkt_reply,
                             krb5_pa_data *data, krb5_kdcpreauth_callbacks cb,
                             krb5_kdcpreauth_rock rock,
                             krb5_kdcpreauth_moddata moddata,
                             krb5_kdcpreauth_verify_respond_fn respond,
                             void *arg)
{
  krb5_error_code retval = KRB5KDC_ERR_PREAUTH_FAILED;
  krb5_keyblock *armor_key = NULL;
  struct config_state cfgs;
  struct verify_state *vs;
  PA_OTP_REQUEST_t *req = NULL;
  char *tmp = NULL;
  unsigned int size;

  cfgs.context = context;
  cfgs.cb = cb;
  cfgs.rock = rock;

  /* Get the FAST armor key */
  armor_key = cb->fast_armor(context, rock);
  if (armor_key == NULL) {
    DEBUGMSG(retval, "No armor key found when verifying padata.");
    (*respond)(arg, EINVAL, NULL, NULL, NULL);
    return;
  }

  /* Decode the request */
  req = int_ber_decode(PA_OTP_REQUEST, (const char*) data->contents,
                       data->length);
  if (!req) {
    DEBUGMSG(retval, "Unable to decode OTP request.");
    (*respond)(arg, EINVAL, NULL, NULL, NULL);
    return;
  }

  /* Decrypt the nonce from the challenge */
  retval = decrypt_encdata(context, armor_key, req, &tmp, &size);
  if (retval != 0) {
    DEBUGMSG(retval, "Unable to decrypt encData.");
    asn_DEF_PA_OTP_REQUEST.free_struct(&asn_DEF_PA_OTP_REQUEST, req, 0);
    (*respond)(arg, retval, NULL, NULL, NULL);
    return;
  }

  /* Verify the nonce or timestamp */
  retval = nonce_verify(context, armor_key, tmp, size);
  if (retval != 0)
    retval = timestamp_verify(context, tmp, size);
  free(tmp);
  if (retval != 0) {
    DEBUGMSG(retval, "Unable to verify nonce or timestamp.");
    asn_DEF_PA_OTP_REQUEST.free_struct(&asn_DEF_PA_OTP_REQUEST, req, 0);
    (*respond)(arg, retval, NULL, NULL, NULL);
    return;
  }

  /* Allocate our misc */
  vs = new0(struct verify_state);
  if (!vs) {
    (*respond)(arg, ENOMEM, NULL, NULL, NULL);
    return;
  }
  vs->respond = respond;
  vs->arg = arg;

  /* Set the flags */
  enc_tkt_reply->flags |= TKT_FLG_PRE_AUTH;
  enc_tkt_reply->flags |= TKT_FLG_HW_AUTH; /* FIXME: Let the OTP method
                                              decide about the HW flag? */

  /* Do the call */
  plugins_verifyRequest((inventory*) moddata,
                        cb->event_context(context, rock), req,
                        get_config, finish_verify, &cfgs, vs, free);
  asn_DEF_PA_OTP_REQUEST.free_struct(&asn_DEF_PA_OTP_REQUEST, req, 0);
}

static krb5_error_code
authhub_server_return_padata(krb5_context context, krb5_pa_data *padata,
                             krb5_data *req_pkt, krb5_kdc_req *request,
                             krb5_kdc_rep *reply, krb5_keyblock *encrypting_key,
                             krb5_pa_data **send_pa_out,
                             krb5_kdcpreauth_callbacks cb,
                             krb5_kdcpreauth_rock rock,
                             krb5_kdcpreauth_moddata moddata,
                             krb5_kdcpreauth_modreq modreq)
{
  krb5_keyblock *armor_key = NULL;

  if (!padata || padata->length == 0)
    return 0;

  /* Get the armor key */
  armor_key = cb->fast_armor(context, rock);
  if (!armor_key) {
    DEBUGMSG(ENOENT, "No armor key found when returning padata.");
    return ENOENT;
  }

  /* Replace the reply key with the FAST armor key.  */
  krb5_free_keyblock_contents(context, encrypting_key);
  return krb5_copy_keyblock_contents(context, armor_key, encrypting_key);
}

static void
authhub_server_free_modreq(krb5_context context,
                           krb5_kdcpreauth_moddata moddata,
                           krb5_kdcpreauth_modreq modreq)
{
}

krb5_error_code
kdcpreauth_authhub_initvt(krb5_context context, int maj_ver, int min_ver,
                          krb5_plugin_vtable vtable)
{
  krb5_kdcpreauth_vtable vt;

  if (maj_ver != 1)
    return KRB5_PLUGIN_VER_NOTSUPP;

  vt = (krb5_kdcpreauth_vtable) vtable;
  vt->name = "authhub";
  vt->pa_type_list = authhub_server_pa_type_list;
  vt->init = authhub_server_init;
  vt->fini = authhub_server_fini;
  vt->flags = authhub_server_flags;
  vt->edata = authhub_server_get_edata;
  vt->verify = authhub_server_verify_padata;
  vt->return_padata = authhub_server_return_padata;
  vt->free_modreq = authhub_server_free_modreq;

  com_err("authhub", 0, "Loaded.");
  return 0;
}

